// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "computations.h"

boolean equal_computation(Computation* left, Computation* right)
{
  if (left == NULL && right == NULL)
    return 1;
  else if (left == NULL)
    return 0;
  else if (right == NULL)
    return 0;

  if (left->type == right->type)
  {
    switch (left->type)
    {
      case COMPUTATION_INPUT_JSAMPLE:
        return (left->component_index == right->component_index);
        break;
      case COMPUTATION_IMMEDIATE_INTEGER:
        return (left->constant == right->constant);
      case COMPUTATION_INTEGER_MULTIPLY:
      case COMPUTATION_INTEGER_ADD:
      case COMPUTATION_INTEGER_RSHIFT:
        return (left->constant == right->constant && equal_computation(left->computation1, right->computation1));
      case COMPUTATION_ADD:
        return (equal_computation(left->computation1, right->computation1) && equal_computation(left->computation2, right->computation2));
    }
  }
  return 0;
}

void print_computation(Computation* computation)
{
  if (computation == NULL)
  {
    printf("nil");
    return;
  }

  switch (computation->type)
  {
    case COMPUTATION_INPUT_JSAMPLE:
      printf("i[%d]", computation->component_index);
      break;
    case COMPUTATION_IMMEDIATE_INTEGER:
      printf("%d", computation->constant);
      break;
    case COMPUTATION_INTEGER_MULTIPLY:
      printf("(");
      print_computation(computation->computation1);
      printf(" * %d)", computation->constant);
      break;
    case COMPUTATION_INTEGER_ADD:
      printf("(");
      print_computation(computation->computation1);
      printf(" + %d)", computation->constant);
      break;
    case COMPUTATION_INTEGER_RSHIFT:
      printf("(");
      print_computation(computation->computation1);
      printf(" >> %d)", computation->constant);
      break;
    case COMPUTATION_ADD:
      printf("+(");
      print_computation(computation->computation1);
      printf(",");
      print_computation(computation->computation2);
      printf(")");
      break;
    default:
      printf("? %d", computation->type);
      break;
  }
  
  if (computation->estimateMax != NULL && computation->estimateMin != NULL)
  {
    printf(":(%ld,%ld)", *computation->estimateMax, *computation->estimateMin);
    if (computation->value != NULL && (*(computation->value) < *(computation->estimateMin) || *(computation->value) > *(computation->estimateMax)))
      printf("#####");
  }
}

// Simplifies a computation using algebraic identities.
// The input clip stride and maximum are used to fix indices into the input JSAMPLE arrays.
Computation* clean_computation(Computation* computation, int input_clip_stride, int input_clip_max)
{
  // Simplify rules:
  // a x + x into (a + 1) x
  // a (b x) into (a b) x
  // Fix indices for input JSAMPLE
  // 1 * a into a
  switch (computation->type)
  {
    case COMPUTATION_INTEGER_MULTIPLY:
      // a (b x) into (a b) x
      if (computation->computation1->type == COMPUTATION_INTEGER_MULTIPLY)
      {
        Computation* result = clean_computation(computation->computation1, input_clip_stride, input_clip_max);
        if (result->type == COMPUTATION_INTEGER_MULTIPLY)
        {
          result->constant *= computation->constant;
          free(computation);
          computation = result;
        }
        else
          computation->computation1 = result;
        return computation;
      }
      else if (computation->computation1->type == COMPUTATION_INPUT_JSAMPLE && computation->constant == 1)
      {
        Computation* result = computation->computation1;
        free(computation);
        return result;
      }
      break;
    case COMPUTATION_ADD:
      // a x + x into (a + 1) x
      // Replace this entire node by an integer multiply node.
      if (computation->computation2->type == COMPUTATION_INPUT_JSAMPLE && computation->computation1->type == COMPUTATION_INTEGER_MULTIPLY && (computation->computation1->computation1 != NULL && computation->computation1->computation1->type == COMPUTATION_INPUT_JSAMPLE && computation->computation1->computation1->component_index == computation->computation2->component_index))
      {
        Computation* result = create_computation(COMPUTATION_INTEGER_MULTIPLY, create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, computation->computation2->component_index), NULL, computation->computation1->constant + 1, 0);
        destroy_computation(computation);
        return result;
      }
      else if (computation->computation1->type == COMPUTATION_INPUT_JSAMPLE && computation->computation2->type == COMPUTATION_INTEGER_MULTIPLY && (computation->computation2->computation1 != NULL && computation->computation2->computation1->type == COMPUTATION_INPUT_JSAMPLE && computation->computation2->computation1->component_index == computation->computation1->component_index))
      {
        Computation* result = create_computation(COMPUTATION_INTEGER_MULTIPLY, create_computation(COMPUTATION_INPUT_JSAMPLE, NULL, NULL, 0, computation->computation1->component_index), NULL, computation->computation2->constant + 1, 0);
        destroy_computation(computation);
        return result;
      }
      break;
    case COMPUTATION_INPUT_JSAMPLE:
      while (computation->component_index < 0)
        computation->component_index += input_clip_stride;
      while (computation->component_index >= input_clip_max)
        computation->component_index -= input_clip_stride;
      break;
    default:
      break;
  }

  // If no rule matches simplify any child computations.
  if (computation->computation1 != NULL)
    computation->computation1 = clean_computation(computation->computation1, input_clip_stride, input_clip_max);
  if (computation->computation2 != NULL)
    computation->computation2 = clean_computation(computation->computation2, input_clip_stride, input_clip_max);

  return computation;
}

void simplify_computation_plane(computation_plane computations)
{
  Computation* working;
  Computation* simplified;
  int i;
  for (i = 0; i < computations.width * computations.height; i++)
  {
    // Simplify the computation.
    working = NULL;
    simplified = computations.computations[i];
    
    // Work out the computation simplification fixed point.
    while (!equal_computation(working, simplified))
    {
      destroy_computation(working); working = simplified;
      simplified = clean_computation(clone_computation(working), computations.input_width, computations.input_width * computations.input_height);
    }
    destroy_computation(working);

    computations.computations[i] = simplified;
  }
}

void evaluate_computation(Computation* computation, long int* max, long int* min, interval_plane current_estimates)
{
  long int max1, min1, temp;
  switch (computation->type)
  {
    case COMPUTATION_INPUT_JSAMPLE:
      *max = current_estimates.max[computation->component_index] == UNION_BUFFER_UNDEFINED ? 255 : current_estimates.max[computation->component_index];
      *min = current_estimates.min[computation->component_index] == UNION_BUFFER_UNDEFINED ? 0 : current_estimates.min[computation->component_index];
      break;
    case COMPUTATION_IMMEDIATE_INTEGER:
      *max = computation->constant;
      *min = computation->constant;
      break;
    case COMPUTATION_INTEGER_MULTIPLY:
      evaluate_computation(computation->computation1, max, min, current_estimates);
      *max *= computation->constant;
      *min *= computation->constant;
      break;
    case COMPUTATION_INTEGER_ADD:
      evaluate_computation(computation->computation1, max, min, current_estimates);
      *max += computation->constant;
      *min += computation->constant;
      break;
    case COMPUTATION_INTEGER_RSHIFT:
      evaluate_computation(computation->computation1, max, min, current_estimates);
      *min >>= computation->constant;
      *max >>= computation->constant;
      break;
    case COMPUTATION_ADD:
      evaluate_computation(computation->computation1, max, min, current_estimates);
      evaluate_computation(computation->computation2, &max1, &min1, current_estimates);
      *max += max1;
      *min += min1;
      break;
  }

  // Fix ordering
  if (*min > *max)
  {
    temp = *min; *min = *max; *max = temp;
  }
}

// Returns the minimum number of binary places which are zero in the LSBs of the computation result.
void minimum_zeros_in_lhs_lsb(Computation* computation, int* current_minimum)
{
  int i;
  switch (computation->type)
  {
    case COMPUTATION_INPUT_JSAMPLE:
      *current_minimum = 0;
      break;
    case COMPUTATION_INTEGER_MULTIPLY:
      minimum_zeros_in_lhs_lsb(computation->computation1, current_minimum);
      i = 1;
      while (i <= 16 && ((computation->constant >> i) << i) == computation->constant)
        i++;
      *current_minimum += i - 1;
      break;
    case COMPUTATION_INTEGER_ADD:
      // Minimum of the embedded expression and the constant's LSBs.
      minimum_zeros_in_lhs_lsb(computation->computation1, current_minimum);
      i = 0;
      while (i <= 16 && ((computation->constant >> i) << i) == computation->constant)
        i++;
      *current_minimum = (*current_minimum < i - 1 ? *current_minimum : i - 1);
      break;
    case COMPUTATION_IMMEDIATE_INTEGER:
      i = 0;
      while (i <= 16 && ((computation->constant >> i) << i) == computation->constant)
        i++;
      *current_minimum = i - 1;
      break;
    case COMPUTATION_INTEGER_RSHIFT:
      minimum_zeros_in_lhs_lsb(computation->computation1, current_minimum);
      *current_minimum = (*current_minimum - computation->constant < 0 ? 0 : *current_minimum - computation->constant);
      break;
    case COMPUTATION_ADD:
      minimum_zeros_in_lhs_lsb(computation->computation1, current_minimum);
      int other_minimum;
      minimum_zeros_in_lhs_lsb(computation->computation2, &other_minimum);
      *current_minimum = (other_minimum < *current_minimum ? other_minimum : *current_minimum);
      break;
  }
}

void reverse_computation(INT64* left_max, INT64* left_min, Computation* computation, interval_plane current_estimates, boolean use_union_buffer, interval_plane last_iteration, boolean debug_output)
{
  long int maximum1, minimum1, maximum2, minimum2;
  long int store_left_max, store_left_min;
  int zeros;

  if (computation == NULL)
    return;

  if (*left_max < *left_min)
  {
    long int temp = *left_max;
    *left_max = *left_min;
    *left_min = temp;
  }

  minimum_zeros_in_lhs_lsb(computation, &zeros);

  if (computation->value != NULL && (*left_min > *(computation->value) || *left_max < *(computation->value)))
  {
    printf("%ld, %ld\n", *left_max, *left_min);
    printf("Inconsistency found!\n");
  }

  if (debug_output)
  {
    printf("%ld %ld on left of ", *left_min, *left_max);
    print_computation(computation);
    printf("\n");
  }

  switch (computation->type)
  {
    case COMPUTATION_INPUT_JSAMPLE:
      // Here, we know some information which might possibly allow us to update the input estimates array.
      if (use_union_buffer)
      {
        // TODO: Move 0, 255 logic higher up.
        if (*left_max > current_estimates.max[computation->component_index])
          current_estimates.max[computation->component_index] = *left_max;
        if (*left_min < current_estimates.min[computation->component_index] || current_estimates.min[computation->component_index] == UNION_BUFFER_UNDEFINED)
          current_estimates.min[computation->component_index] = *left_min;
      }
      else
      {
        /* Update the minima and maxima directly. */
        if (*left_max < current_estimates.max[computation->component_index])
        {
          if (*left_max < current_estimates.min[computation->component_index])
          {
            printf("MAXInconsistency: (%ld, %ld) for (%d, %d)\n", *left_max, *left_min, current_estimates.max[computation->component_index], current_estimates.min[computation->component_index]);
          }
          else
          {
            current_estimates.max[computation->component_index] = *left_max;
          }
        }
        if (*left_min > current_estimates.min[computation->component_index])
        {
          if (*left_min > current_estimates.max[computation->component_index])
          {
            printf("MINInconsistency: (%ld, %ld) for (%d, %d)\n", *left_max, *left_min, current_estimates.max[computation->component_index], current_estimates.min[computation->component_index]);
          }
          else
          {
            current_estimates.min[computation->component_index] = *left_min;
          }
        }
      }
      break;
    case COMPUTATION_IMMEDIATE_INTEGER:
      // This should never be reached.
      printf("An error occurred during equation re-arrangement. Immediate integer node was reached.\n");
      break;

    case COMPUTATION_INTEGER_MULTIPLY:
      if (*left_max < 0)
      {
        *left_max = ((*left_max % computation->constant == 0) || (*left_max / computation->constant == 0)) ? *left_max / computation->constant : (*left_max / computation->constant) - 1;
      }
      else
      {
        *left_max = *left_max / computation->constant;
      }

      if (*left_min < 0)
      {
        *left_min = *left_min / computation->constant;
      }
      else
      {
        *left_min = ((*left_min % computation->constant == 0) || (*left_min / computation->constant == 0)) ? *left_min / computation->constant : (*left_min / computation->constant) + 1;
      }
      reverse_computation(left_max, left_min, computation->computation1, current_estimates, use_union_buffer, last_iteration, debug_output);
      break;
    case COMPUTATION_INTEGER_ADD:
      *left_max -= computation->constant;
      *left_min -= computation->constant;
      reverse_computation(left_max, left_min, computation->computation1, current_estimates, use_union_buffer, last_iteration, debug_output);
      break;
    case COMPUTATION_INTEGER_RSHIFT:
      // To work out the set of possible error values, get the minimum number of zeros possible in the LSBs of the computation.
      minimum_zeros_in_lhs_lsb(computation->computation1, &zeros);

      *left_max = (*left_max * (1 << computation->constant)) + ((((1 << computation->constant) - 1) >> zeros) << zeros);
      *left_min = (*left_min * (1 << computation->constant)) + 0;

      reverse_computation(left_max, left_min, computation->computation1, current_estimates, use_union_buffer, last_iteration, debug_output);
      break;
    case COMPUTATION_ADD:
      // Estimate each branch.

      // Try to tighten bound in the left parameter of add.
      evaluate_computation(computation->computation2, &maximum2, &minimum2, last_iteration);
      store_left_max = *left_max; store_left_min = *left_min;
      *left_max -= minimum2; *left_min -= maximum2;
      reverse_computation(left_max, left_min, computation->computation1, current_estimates, use_union_buffer, last_iteration, debug_output);
      *left_max = store_left_max; *left_min = store_left_min;

      // Try to tighten bound in the right parameter of add.
      evaluate_computation(computation->computation1, &maximum1, &minimum1, last_iteration);
      *left_max -= minimum1; *left_min -= maximum1;
      reverse_computation(left_max, left_min, computation->computation2, current_estimates, use_union_buffer, last_iteration, debug_output);
      // TODO: Next line is not referenced.
      *left_max = store_left_max; *left_min = store_left_min;
      break;
  }
}
