// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "rjpeg.h"
#include "cspace.h"

inverse_ycc_rgb_table get_inverse_ycc_rgb_table()
{
  // Check if the pre-generated tables exist. If not, regenerate them.
  // TODO: Should the user be able to specify this file name?
  FILE* table_file = fopen(INVERSE_YCC_RGB_TABLE_FILE_NAME, "r"); 
  if (table_file != NULL)
  {
    fclose(table_file);
    return load_inverse_ycc_rgb_table(INVERSE_YCC_RGB_TABLE_FILE_NAME);
  }
  else
    return generate_inverse_ycc_rgb_table(INVERSE_YCC_RGB_TABLE_FILE_NAME);
}

void write_UINT32(FILE* file, const UINT32 value)
{
  fputc((int)((value & 0x000000FF) >> 0), file);
  fputc((int)((value & 0x0000FF00) >> 8), file);
  fputc((int)((value & 0x00FF0000) >> 16), file);
  fputc((int)((value & 0xFF000000) >> 24), file);
}

UINT32 read_UINT32(FILE* file)
{
  UINT32 result;
  result = fgetc(file);
  result += fgetc(file) << 8;
  result += fgetc(file) << 16;
  result += fgetc(file) << 24;
  return result;
}

set_plane reverse_colour_space_transform(packed_image_data rgb_image, inverse_ycc_rgb_table ycc_rgb_table)
{
  // The result of the colour space transform will not have chroma subsampled.
  set_plane component_data = create_set_plane(rgb_image);
  unsigned int x, y;
  unsigned char r, g, b;

  // Iterate over the original image, looking up each RGB tuples in the inverted table.
  // Set the interval_planes values to include the union of the values over all possibilities.
  for (y = 0; y < rgb_image.height; y++)
  {
    for (x = 0; x < rgb_image.width; x++)
    {
      r = rgb_image.samples[((y * rgb_image.width) + x) * 3 + 0];
      g = rgb_image.samples[((y * rgb_image.width) + x) * 3 + 1];
      b = rgb_image.samples[((y * rgb_image.width) + x) * 3 + 2];

      UINT32* component_list = ycc_rgb_table.conversion_table[r * 256 * 256 + g * 256 + b];

      if (component_list == NULL)
        printf("An invalid colour tuple was present in the input. This image could not have resulted from JPEG decompression with IJG.");

      component_data.sets[y * rgb_image.width + x] = component_list;
    }
  }

  return component_data;
}

void calculate_conversion_tabs(inverse_ycc_rgb_table* rgb_to_component)
{
  INT32 i, x;
  rgb_to_component->cr_r_tab = (INT16*)malloc(256 * sizeof(INT16));
  rgb_to_component->cb_b_tab = (INT16*)malloc(256 * sizeof(INT16));
  rgb_to_component->cr_g_tab = (INT32*)malloc(256 * sizeof(INT32));
  rgb_to_component->cb_g_tab = (INT32*)malloc(256 * sizeof(INT32));

  // Generate the tables to convert YCbCr into RGB values.
  for (i = 0, x = -CENTRE_JSAMPLE; i <= MAX_JSAMPLE; i++, x++)
  {
    rgb_to_component->cr_r_tab[i] = (INT16)RIGHT_SHIFT(FIX(1.40200) * x + ONE_HALF, SCALEBITS);
    rgb_to_component->cb_b_tab[i] = (INT16)RIGHT_SHIFT(FIX(1.77200) * x + ONE_HALF, SCALEBITS);
    rgb_to_component->cr_g_tab[i] = (-FIX(0.71414)) * x;
    rgb_to_component->cb_g_tab[i] = (-FIX(0.34414)) * x + ONE_HALF;
  }
}

// Generates tables required for reversing the YCbCr to RGB conversion.
inverse_ycc_rgb_table generate_inverse_ycc_rgb_table(const char* table_file)
{
  int i, j;
  int y, cb, cr, r, g, b;
  INT32 r_full, g_full, b_full;
  UINT32* list;

  // Tables as generated in IJG.
  inverse_ycc_rgb_table rgb_to_component = create_inverse_ycc_rgb_table();
  calculate_conversion_tabs(&rgb_to_component);

  // Invert the look-up, storing sets of possible YCbCr tuples for each RGB input.
  for ( i = 0; i < 256 * 256 * 256; i++)
    rgb_to_component.conversion_table[i] = NULL;
  for (y = 0; y < 256; y++)
  {
    for (cb = 0; cb < 256; cb++)
    {
      for (cr = 0; cr < 256; cr++)
      {
        r_full = y + rgb_to_component.cr_r_tab[cr];
        g_full = y + ((INT16)RIGHT_SHIFT(rgb_to_component.cb_g_tab[cb] + rgb_to_component.cr_g_tab[cr], SCALEBITS));
        b_full = y + rgb_to_component.cb_b_tab[cb];
        r = r_full > MAX_JSAMPLE ? MAX_JSAMPLE : (r_full < 0 ? 0 : r_full);
        g = g_full > MAX_JSAMPLE ? MAX_JSAMPLE : (g_full < 0 ? 0 : g_full);
        b = b_full > MAX_JSAMPLE ? MAX_JSAMPLE : (b_full < 0 ? 0 : b_full);
        list = rgb_to_component.conversion_table[r * 256 * 256 + g * 256 + b];

        // Allocate a new list if required.
        if (list == NULL)
        {
          list = rgb_to_component.conversion_table[r * 256 * 256 + g * 256 + b] = (UINT32*)malloc(LIST_START_SIZE * sizeof(UINT32));
          list[0] = LIST_START_SIZE;
          for (i = 1; i < LIST_START_SIZE; i++)
            list[i] = LIST_SENTINEL32;
        }

        // Append this item to the list, resizing if required.
        i = 1;
        while (list[i] != LIST_SENTINEL32)
        {
          i++;
          if (i == list[0])
          {
            list = rgb_to_component.conversion_table[r * 256 * 256 + g * 256 + b] = (UINT32*)realloc(list, (list[0] + LIST_START_SIZE) * sizeof(UINT32));
            list[0] = list[0] + LIST_START_SIZE;
            for (j = i; j < list[0]; j++)
              list[j] = LIST_SENTINEL32;
            break;
          }
        }

        list[i] = y * 256 * 256 + cb * 256 + cr;
      }
    }
  }

  // Write the list to a file, and return it.
  FILE* table_write = fopen(table_file, "wb");
  UINT32 value_count;
  for (r = 0; r < 256; r++)
  {
    for (g = 0; g < 256; g++)
    {
      for (b = 0; b < 256; b++)
      {
        list = rgb_to_component.conversion_table[r * 256 * 256 + g * 256 + b];
        if (list == NULL)
        {
          write_UINT32(table_write, 0);
        }
        else
        {
          // Work out the length of the list.
          value_count = 0;
          for (i = 1; i < list[0]; i++)
          {
            if (list[i] != LIST_SENTINEL32)
              value_count++;
            else
              break;
          }

          // Output the number of values and then a list.
          write_UINT32(table_write, value_count);
          for (i = 1; i < value_count + 1; i++)
            write_UINT32(table_write, list[i]);
        }
      }
    }
  }
  fclose(table_write);

  return rgb_to_component;
}

// Load tables for colour space conversion from a file.
inverse_ycc_rgb_table load_inverse_ycc_rgb_table(const char* table_file)
{
  FILE* table_read = fopen(table_file, "rb");
  UINT32 value_count, i;
  UINT32* list;
  inverse_ycc_rgb_table rgb_to_component = create_inverse_ycc_rgb_table();
  calculate_conversion_tabs(&rgb_to_component);
  
  int r, g, b;
  for (r = 0; r < 256; r++)
  {
    for (g = 0; g < 256; g++)
    {
      for (b = 0; b < 256; b++)
      {
        list = NULL;
        value_count = read_UINT32(table_read);
        if (value_count == 0)
        {
          rgb_to_component.conversion_table[r * 256 * 256 + g * 256 + b] = NULL;
        }
        else
        {
          list = rgb_to_component.conversion_table[256 * 256 * r + 256 * g + b] = (UINT32*)malloc((value_count + 1) * sizeof(UINT32));
          list[0] = value_count + 1;
          for (i = 0; i < value_count; i++)
            list[i + 1] = read_UINT32(table_read);
        }
      }
    }
  }

  return rgb_to_component;
}
