// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "data.h"

quantization_table create_quantization_table()
{
  return (quantization_table)malloc(64 * sizeof(UINT16));
}

void destroy_quantization_table(quantization_table value)
{
  free(value);
}

packed_image_data create_packed_image_data(UINT16 width, UINT16 height)
{
  packed_image_data result;
  result.width = width;
  result.height = height;
  result.samples = NULL;
  return result;
}

void destroy_packed_image_data(packed_image_data data)
{
  free(data.samples);
}

inverse_ycc_rgb_table create_inverse_ycc_rgb_table()
{
  inverse_ycc_rgb_table result;
  result.conversion_table = malloc(256 * 256 * 256 * sizeof(UINT32*));
  result.cb_g_tab = result.cr_g_tab = result.cb_b_tab = result.cr_r_tab = NULL;
  return result;
}

void destroy_inverse_ycc_rgb_table(inverse_ycc_rgb_table table)
{
  int r, g, b;
  for (r = 0; r < 256; r++)
    for (g = 0; g < 256; g++)
      for (b = 0; b < 256; b++)
        if (table.conversion_table[r * 256 * 256 + g * 256 + b] != NULL)
          free(table.conversion_table[r * 256 * 256 + g * 256 + b]);
  free(table.conversion_table);

  free(table.cr_r_tab);
  free(table.cb_b_tab);
  free(table.cr_g_tab);
  free(table.cb_g_tab);
}

interval_plane create_interval_plane(UINT16 width, UINT16 height)
{
  interval_plane result;
  result.min = (INT16*)malloc(width * height * sizeof(INT16));
  result.max = (INT16*)malloc(width * height * sizeof(INT16));
  result.width = width;
  result.height = height;

  return result;
}

void destroy_interval_plane(interval_plane plane)
{
  free(plane.min);
  free(plane.max);
}

interval_component_image_data create_interval_component_image_data(packed_image_data original, boolean chroma_subsampling)
{
  interval_component_image_data result;
  result.width = original.width;
  result.height = original.height;
  result.y = create_interval_plane(original.width, original.height);
  if (chroma_subsampling)
  {
    // TODO: Deal with non-multiple of two sizes.
    result.cb = create_interval_plane(original.width / 2, original.height / 2);
    result.cr = create_interval_plane(original.width / 2, original.height / 2);
  }
  else
  {
    result.cb = create_interval_plane(original.width, original.height);
    result.cr = create_interval_plane(original.width, original.height);
  }

  return result;
}

void set_maximum_chroma_intervals(interval_component_image_data data)
{
  UINT16 x, y;
  for (y = 0; y < data.cb.height; y++)
  {
    for (x = 0; x < data.cb.width; x++)
    {
      data.cb.min[y * data.cb.width + x] = 0;
      data.cb.max[y * data.cb.width + x] = 255;
      data.cr.min[y * data.cb.width + x] = 0;
      data.cr.max[y * data.cb.width + x] = 255;
    }
  }
}

void copy_luma_data_into_component_data(set_plane luma_source, interval_component_image_data luma_destination, const UINT8 luma_shift)
{
  UINT16 x, y, set_index;
  INT16 min, max, sample;
  if (luma_source.width != luma_destination.y.width || luma_source.height != luma_destination.y.height)
    printf("Luma source/destination size mismatch.\n");
  for (y = 0; y < luma_destination.height; y++)
  {
    for (x = 0; x < luma_destination.width; x++)
    {
      min = max = -1;
      for (set_index = 1; set_index < luma_source.sets[y * luma_destination.width + x][0]; set_index++)
      {
        if (luma_source.sets[y * luma_source.width + x][set_index] == LIST_SENTINEL32)
          break;
        sample = ((luma_source.sets[y * luma_source.width + x][set_index] >> luma_shift) & 0x0000FF);
  
        // TODO: Needs to be changed when signed values stored in the sets.
        if (sample < min || min == -1)
          min = sample;
        if (sample > max || max == -1)
          max = sample;
      }

      luma_destination.y.min[y * luma_destination.width + x] = min;
      luma_destination.y.max[y * luma_destination.width + x] = max;
    }
  }
}

void destroy_interval_component_image_data(interval_component_image_data data)
{
  destroy_interval_plane(data.y);
  destroy_interval_plane(data.cb);
  destroy_interval_plane(data.cr);
}

set_plane create_set_plane(packed_image_data original)
{
  set_plane result;
  result.width = original.width;
  result.height = original.height;
  result.sets = (UINT32**)malloc(original.width * original.height * sizeof(UINT32*));

  return result;
}

// Destroys a set plane, including the contents of sets.
void destroy_set_plane(set_plane data)
{
  UINT16 y, x;

  // Free each set's contents.
  for (y = 0; y < data.height; y++)
  {
    for (x = 0; x < data.width; x++)
    {
      free(data.sets[y * data.width + x]);
    }
  }

  free(data.sets);
}

block_interval_component_image_data create_block_interval_component_image_data(interval_component_image_data data)
{
  // TODO: Assuming width and height are multiples of eight.
  block_interval_component_image_data result;
  UINT16 bx;
  UINT16 by;
  UINT16 bi;
  result.blocks_width = data.width / 8;
  result.blocks_height = data.height / 8;
  result.y = create_interval_plane(data.y.width, data.y.height);
  result.cb = create_interval_plane(data.cb.width, data.cb.height);
  result.cr = create_interval_plane(data.cr.width, data.cr.height);
  
  // Copy the data so that the interval planes contain a sequence of 8 * 8 block samples.
  for (by = 0; by < result.y.height / 8; by++)
  {
    for (bx = 0; bx < result.y.width / 8; bx++)
    {
      for (bi = 0; bi < 64; bi++)
      {
        result.y.min[64 * (bx + by * (result.y.width / 8)) + bi] = data.y.min[(by * 8 * result.y.width) + (bx * 8) + ((bi / 8) * result.y.width) + (bi % 8)];
        result.y.max[64 * (bx + by * (result.y.width / 8)) + bi] = data.y.max[(by * 8 * result.y.width) + (bx * 8) + ((bi / 8) * result.y.width) + (bi % 8)];
      }
    }
  }

  // Same for the chroma channels.
  for (by = 0; by < result.cb.height / 8; by++)
  {
    for (bx = 0; bx < result.cb.width / 8; bx++)
    {
      for (bi = 0; bi < 64; bi++)
      {
        result.cb.min[64 * (bx + by * (result.cb.width / 8)) + bi] = data.cb.min[(by * 8 * result.cb.width) + (bx * 8) + ((bi / 8) * result.cb.width) + (bi % 8)];
        result.cb.max[64 * (bx + by * (result.cb.width / 8)) + bi] = data.cb.max[(by * 8 * result.cb.width) + (bx * 8) + ((bi / 8) * result.cb.width) + (bi % 8)];
        result.cr.min[64 * (bx + by * (result.cr.width / 8)) + bi] = data.cr.min[(by * 8 * result.cr.width) + (bx * 8) + ((bi / 8) * result.cr.width) + (bi % 8)];
        result.cr.max[64 * (bx + by * (result.cr.width / 8)) + bi] = data.cr.max[(by * 8 * result.cr.width) + (bx * 8) + ((bi / 8) * result.cr.width) + (bi % 8)];
      }
    }
  }

  return result;
}

void destroy_block_interval_component_image_data(block_interval_component_image_data data)
{
  destroy_interval_plane(data.y);
  destroy_interval_plane(data.cb);
  destroy_interval_plane(data.cr);
}

computation_plane create_computation_plane(set_plane data, UINT16 input_width, UINT16 input_height)
{
  computation_plane result;
  result.computations = (Computation**)malloc(data.width * data.height * sizeof(Computation*));
  result.width = data.width;
  result.height = data.height;
  result.input_width = input_width;
  result.input_height = input_height;

  return result;
}

void destroy_computation_plane(computation_plane computations)
{
  int i;
  for (i = 0; i < computations.width * computations.height; i++)
    destroy_computation(computations.computations[i]);
  free(computations.computations);
}

Computation* clone_computation(Computation* computation)
{
  if (computation == NULL)
    return NULL;
  Computation* result = (Computation*)malloc(sizeof(Computation));
  memcpy(result, computation, sizeof(Computation));
  result->computation1 = clone_computation(computation->computation1);
  result->computation2 = clone_computation(computation->computation2);
  return result;
}

Computation* create_computation(ComputationType type, Computation* c1, Computation* c2, int constant, int component_index)
{
  Computation* result = (Computation*)malloc(sizeof(Computation));
  result->type = type;
  result->computation1 = c1;
  result->computation2 = c2;
  result->constant = constant;
  result->component_index = component_index;
  result->is_known = 0;
  result->value = NULL;
  result->estimateMax = NULL;
  result->estimateMin = NULL;
  return result;
}

void destroy_computation(Computation* computation)
{
  if (computation == NULL)
    return;
  if (computation->computation1 != NULL)
    destroy_computation(computation->computation1);
  if (computation->computation2 != NULL)
    destroy_computation(computation->computation2);
  if (computation->value != NULL)
    free(computation->value);
  if (computation->estimateMax != NULL)
    free(computation->estimateMax);
  if (computation->estimateMin != NULL)
    free(computation->estimateMin);
  free(computation);
}
