// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "quantize.h"

void get_quantization_matrix(UINT8 quality_factor, UINT16* luma_table, UINT16* chroma_table)
{
  static const unsigned int std_luminance_quant_tbl[8 * 8] = {
    16,  11,  10,  16,  24,  40,  51,  61,
    12,  12,  14,  19,  26,  58,  60,  55,
    14,  13,  16,  24,  40,  57,  69,  56,
    14,  17,  22,  29,  51,  87,  80,  62,
    18,  22,  37,  56,  68, 109, 103,  77,
    24,  35,  55,  64,  81, 104, 113,  92,
    49,  64,  78,  87, 103, 121, 120, 101,
    72,  92,  95,  98, 112, 100, 103,  99};
  static const unsigned int std_chrominance_quant_tbl[8 * 8] = {
    17,  18,  24,  47,  99,  99,  99,  99,
    18,  21,  26,  66,  99,  99,  99,  99,
    24,  26,  56,  99,  99,  99,  99,  99,
    47,  66,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99,
    99,  99,  99,  99,  99,  99,  99,  99};

  // Scale the quality.
  UINT16 scaled_quality = (quality_factor < 50 ? (5000 / quality_factor) : (200 - quality_factor * 2));
  UINT8 bi;
  long temp;

  for (bi = 0; bi < 64; bi++)
  {
    temp = ((long)std_luminance_quant_tbl[bi] * scaled_quality + 50L) / 100L;
    luma_table[bi] = temp <= 0L ? 1 : (temp > 255L ? 255 : (UINT16)(temp));

    temp = ((long)std_chrominance_quant_tbl[bi] * scaled_quality + 50L) / 100L;
    chroma_table[bi] = temp <= 0L ? 1 : (temp > 255L ? 255 : (UINT16)(temp));
  }
}

// The possible map is 64 * 2 = 128 bitmaps, arranged in low/high, low/high, ... order.
// For new possible maps, the elements should be set to zero.
void quantize_interval_plane(interval_plane dct_intervals, UINT64* possible_map)
{
  UINT16 x, y;
  UINT8 bi, qval;
  INT32 cmin, cmax;
  UINT64 pmap_highhigh, pmap_high, pmap_low;

  // We are provided with intervals on DCT coefficients over the whole image.
  // Work out the maximum possible quantization factor for the values.
  // TODO: Not going to get same file if the quantization was inefficient.
  // TODO: Assuming block size is eight by eight.
  for (bi = 0; bi < 64; bi++)
  {
    pmap_low = possible_map[bi * 3];
    pmap_high = possible_map[bi * 3 + 1];
    pmap_highhigh = possible_map[bi * 3 + 2];

    for (y = 0; y < dct_intervals.height / 8; y++)
    {
      for (x = 0; x < dct_intervals.width / 8; x++)
      {
        cmin = dct_intervals.min[(y * dct_intervals.width / 8 + x) * 64 + bi];
        cmax = dct_intervals.max[(y * dct_intervals.width / 8 + x) * 64 + bi];

        // Test each quantization value.
        for (qval = 2; qval < 192; qval++)
        {
          // Check for no quantized values in the interval, and not being exactly on the interval.
          if ((cmax / qval) == (cmin / qval) && (cmin < 0 ? cmax : cmin) % qval != 0 && !(cmax >= 0 && cmin <= 0))
          {
            if (qval > 127)
              pmap_highhigh |= (1L << (qval - 128));
            else if (qval > 63)
              pmap_high |= (1L << (qval - 64));
            else
              pmap_low |= (1L << (qval));
          }
        }
      }
    }

    possible_map[bi * 3] = pmap_low;
    possible_map[bi * 3 + 1] = pmap_high;
    possible_map[bi * 3 + 2] = pmap_highhigh;
  }
}

void display_possibles_map(UINT64 low, UINT64 high, UINT64 highhigh)
{
  UINT8 qval;
  for (qval = 0; qval < 192; qval++)
  {
    if (qval > 127)
    {
      if ((highhigh & (1 << (qval - 128))) != 0)
        printf("0");
      else
        printf("1");
    }
    else if (qval > 63)
    {
      if ((high & (1 << (qval - 64))) != 0)
        printf("0");
      else
        printf("1");
    }
    else
    {
      if ((low & (1 << qval)) != 0)
        printf("0");
      else
        printf("1");
    }
  }
  printf("\n");
}

void perform_quantization(block_interval_component_image_data image, quantization_table luma_q_table, quantization_table chroma_q_table)
{
  UINT8 bx, by, bi;
  INT32 min, max, qval;

  /*printf("  Quantizing using tables:\n ");
  for (bi = 0; bi < 64; bi++)
  {
    if (bi % 8 == 0)
      printf("\n  ");
    printf("%d ", (INT32)luma_q_table[bi]);
  }

  printf("\n  ");
  for (bi = 0; bi < 64; bi++)
  {
    if (bi % 8 == 0)
      printf("\n  ");
    printf("%d ", (INT32)chroma_q_table[bi]);
  }
  printf("\n");*/

  // Luma quantization
  for (by = 0; by < image.y.height / 8; by++)
  {
    for (bx = 0; bx < image.y.width / 8; bx++)
    {
      for (bi = 0; bi < 64; bi++)
      {
        qval = (INT32)luma_q_table[bi];
        min = (INT32)image.y.min[(by * image.y.width / 8 + bx) * 64 + bi];
        max = (INT32)image.y.max[(by * image.y.width / 8 + bx) * 64 + bi];

        if (min < 0)
          image.y.min[(by * image.y.width / 8 + bx) * 64 + bi] = (min / qval);
        else
          image.y.min[(by * image.y.width / 8 + bx) * 64 + bi] = (min % qval == 0 ? (min / qval) : ((min + qval) / qval));

        if (max < 0)
          image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] = (max % qval == 0 ? (max / qval) : ((max - qval) / qval));
        else
          image.y.max[(by * image.y.width / 8 + bx) * 64 + bi] = (max / qval);
      }
    }
  }

  // Chroma quantization.
  for (by = 0; by < image.cb.height / 8; by++)
  {
    for (bx = 0; bx < image.cb.width / 8; bx++)
    {
      for (bi = 0; bi < 64; bi++)
      {
        qval = (INT32)chroma_q_table[bi];

        // The minimum value rounds away from zero. The maximum values rounds towards zero.
        min = (INT32)image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi];
        max = (INT32)image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi];

        if (min < 0)
          image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi] = (min / qval);
        else
          image.cb.min[(by * image.cb.width / 8 + bx) * 64 + bi] = (min % qval == 0 ? (min / qval) : ((min + qval) / qval));

        if (max < 0)
          image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] = (max % qval == 0 ? (max / qval) : ((max - qval) / qval));
        else
          image.cb.max[(by * image.cb.width / 8 + bx) * 64 + bi] = (max / qval);

        // Same rounding scheme for Cr.
        min = (INT32)image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi];
        max = (INT32)image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi];

        if (min < 0)
          image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi] = min / qval;
        else
          image.cr.min[(by * image.cr.width / 8 + bx) * 64 + bi] = (min % qval == 0 ? (min / qval) : ((min + qval) / qval));

        if (max < 0)
          image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi] = (max % qval == 0 ? (max / qval) : ((max - qval) / qval));
        else
          image.cr.max[(by * image.cr.width / 8 + bx) * 64 + bi] = (max / qval);
      }
    }
  }
}

// Work out the quality factor for the image, and the associated quantization tables. The chosen quantization tables will be output in lowest_qf_luma_table and lowest_qf_chroma_table.
boolean quantize_intervals(block_interval_component_image_data image, quantization_table* lowest_qf_luma_table, quantization_table* lowest_qf_chroma_table, UINT8* determined_qf, UINT8 qf_lower_limit)
{
  UINT8 qf, bi;
  UINT16 qval_y, qval_c;
  boolean works, quantization_possible;
  UINT64* possible_map_luma = (UINT64*)malloc(192 * sizeof(UINT64));
  UINT64* possible_map_chroma = (UINT64*)malloc(192 * sizeof(UINT64));
  quantization_table luma_table = create_quantization_table();
  quantization_table chroma_table = create_quantization_table();
  (*lowest_qf_luma_table) = NULL;
  (*lowest_qf_chroma_table) = NULL;
  quantization_possible = FALSE;

  // Clear the possibilities maps.
  memset(possible_map_luma, 0, sizeof(UINT64) * 192);
  memset(possible_map_chroma, 0, sizeof(UINT64) * 192);

  // Work out the possible quantization values for each coefficient. Chroma possibilities are propagated on to Cr after Cb calculations.
  quantize_interval_plane(image.y, possible_map_luma);
  quantize_interval_plane(image.cb, possible_map_chroma);
  quantize_interval_plane(image.cr, possible_map_chroma);

  // TODO: Replace with a binary search.
  // Search through the possible quality factors/quantization matrices.
  for (qf = 1; qf < 100; qf++)
  {
    get_quantization_matrix(qf, luma_table, chroma_table);
    works = TRUE;
    for (bi = 0; bi < 64; bi++)
    {
      qval_y = luma_table[bi];
      qval_c = chroma_table[bi];

      if (qval_y > 127)
      {
        if ((possible_map_luma[bi * 3 + 2] & (1L << (qval_y - 128))) != 0)
        {
          works = FALSE;
          break;
        }
      }
      else if (qval_y > 63)
      {
        if ((possible_map_luma[bi * 3 + 1] & (1L << (qval_y - 64))) != 0)
        {
          works = FALSE;
          break;
        }
      }
      else
      {
        if ((possible_map_luma[bi * 3] & (1L << qval_y)) != 0)
        {
          works = FALSE;
          break;
        }
      }

      if (qval_c > 127)
      {
        if ((possible_map_chroma[bi * 3 + 2] & (1L << (qval_c - 128))) != 0)
        {
          works = FALSE;
          break;
        }
      }
      else if (qval_c > 63)
      {
        if ((possible_map_chroma[bi * 3 + 1] & (1L << (qval_c - 64))) != 0)
        {
          works = FALSE;
          break;
        }
      }
      else
      {
        if ((possible_map_chroma[bi * 3] & (1L << qval_c)) != 0)
        {
          works = FALSE;
          break;
        }
      }
    }

    if (works == TRUE)
    {
      if ((*lowest_qf_luma_table) == NULL && (*lowest_qf_chroma_table) == NULL && qf > qf_lower_limit)
      {
        (*lowest_qf_luma_table) = create_quantization_table();
        (*lowest_qf_chroma_table) = create_quantization_table();
        memcpy(*lowest_qf_luma_table, luma_table, sizeof(UINT16) * 64);
        memcpy(*lowest_qf_chroma_table, chroma_table, sizeof(UINT16) * 64);
        *determined_qf = qf;
        quantization_possible = TRUE;
      }
    }
  }

  // Clean up temporary data.
  destroy_quantization_table(luma_table);
  destroy_quantization_table(chroma_table);
  free(possible_map_luma);
  free(possible_map_chroma);

  // Return FALSE and don't try to quantize with uninitialized data.
  if (!quantization_possible)
    return quantization_possible;

  // Perform quantization on the intervals.
  perform_quantization(image, *lowest_qf_luma_table, *lowest_qf_chroma_table);

  // We have performed quantization so this will always return TRUE.
  return quantization_possible;
}
