// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "rjpeg.h"
#include "cspace.h"
#include "computations.h"
#include "unsmooth.h"
#include "reverseidct.h"
#include "quantize.h"
#include "jpegout.h"

//#define PROGRESS_OUTPUT

int main(int argc, const char* argv[])
{
  if (argc < 2)
  {
    printf("Usage: %s image.ppm [image2.ppm [...]]\nRecompress each specified image\n", argv[0]);
    return 0;
  }

  // Get the table for colour space conversion reversal.
  inverse_ycc_rgb_table inverse_ycc_rgb_table;
  progress_output("Preparing the colour space transform table\n");
  inverse_ycc_rgb_table = get_inverse_ycc_rgb_table();

  UINT16 batch_index;
  char* file_name_buffer = (char*)malloc(255 * sizeof(char));
  for (batch_index = 1; batch_index < argc; batch_index++)
  {
    computation_plane smoothing_computations;
    packed_image_data image;
    set_plane upsampled_component_planes;
    interval_component_image_data downsampled_chroma_estimate;
    block_interval_component_image_data block_data;
    block_interval_component_image_data refined_block_data;
    quantization_table* luma_table;
    quantization_table* chroma_table;
    block_status_summary y_block_status, cb_block_status, cr_block_status;
    UINT8 determined_qf, qf_lower_limit;
    boolean all_blocks_possible, quantization_possible;
    UINT16 iterations, iteration_limit;
  
    // Load the image.
    image = load_rgb_ppm(argv[batch_index]);
  
    // Do the colour space conversion.
    progress_output("Reversing the colour space transform\n");
    upsampled_component_planes = reverse_colour_space_transform(image, inverse_ycc_rgb_table);
  
    // Store our current estimate of the downsampled chroma channels.
    downsampled_chroma_estimate = create_interval_component_image_data(image, TRUE);
    set_maximum_chroma_intervals(downsampled_chroma_estimate);
  
    // Get a set of computations describing the upsampling operation.
    progress_output("Determining computations for chroma upsampling\n");
    smoothing_computations = get_smoothing_computations_plane(upsampled_component_planes);
    simplify_computation_plane(smoothing_computations);
  
    // Prepare quantization table pointers.
    luma_table = (quantization_table*)malloc(sizeof(quantization_table));
    chroma_table = (quantization_table*)malloc(sizeof(quantization_table));
  
    iteration_limit = 1;
    iterations = 0;
    qf_lower_limit = 0;
    while (TRUE)
    {
      // Solve in all ways to derive intervals for the input to chroma upsampling, then copy the luma across.
      progress_output("Recompression iteration %d\n", iterations + 1);
      progress_output(" Unsmoothing chroma channels\n");
      calculate_input_intervals(smoothing_computations, downsampled_chroma_estimate.cb, upsampled_component_planes, SHIFT_CB);
      calculate_input_intervals(smoothing_computations, downsampled_chroma_estimate.cr, upsampled_component_planes, SHIFT_CR);
      copy_luma_data_into_component_data(upsampled_component_planes, downsampled_chroma_estimate, SHIFT_Y);
    
      // Re-order the data to be in 64-element blocks and reverse the IDCT.
      progress_output(" Reversing the inverse DCT\n");
      block_data = create_block_interval_component_image_data(downsampled_chroma_estimate);
      reverse_idct_interval(block_data);
  
      // Derive quantization matrices, and refine intervals.
      progress_output(" Quantizing the DCT coefficient intervals\n");
      quantization_possible = quantize_intervals(block_data, luma_table, chroma_table, &determined_qf, qf_lower_limit);
    
      // If quantization was impossible, leave the loop now.
      if (!quantization_possible)
      {
        destroy_block_interval_component_image_data(block_data);
        break;
      }
  
      // Search the possibilities for chroma coefficients, to refine them.
      progress_output(" Refining luma and chroma ranges\n");
      refined_block_data = create_block_interval_component_image_data(downsampled_chroma_estimate);

      sprintf(file_name_buffer, "%s.additional.txt", argv[batch_index]);
      all_blocks_possible = refine_luma_and_chroma(block_data, refined_block_data, *luma_table, *chroma_table, &downsampled_chroma_estimate, smoothing_computations, inverse_ycc_rgb_table, image, &y_block_status, &cb_block_status, &cr_block_status, iterations == iteration_limit, file_name_buffer);
      destroy_block_interval_component_image_data(block_data);
  
      // TODO: Replace with progress test.
      if (iterations == iteration_limit)
        break;
    
      // refined_block_data contains the refined quantized DCT coefficients, which we don't need if we are doing another pass.
      destroy_block_interval_component_image_data(refined_block_data);
  
      if (all_blocks_possible)
      {
        // Upsample the new chroma planes, and use the result to refine the initial luma and chroma possibilities.
        progress_output(" Refining the upsampled chroma ranges\n");
        refine_upsampled_chroma(downsampled_chroma_estimate, upsampled_component_planes, smoothing_computations);
      }
      else
      {
        if (determined_qf == 100)
        {
          // Nothing more can be done for this image, so stop on the next iteration.
          iteration_limit = iterations + 1;
        }
        else
        {
          qf_lower_limit = determined_qf;
          iteration_limit++;
        }
      }
  
      // Destroy the quantization tables from this iteration.
      destroy_quantization_table(*luma_table);
      destroy_quantization_table(*chroma_table);
  
      iterations++;
    }
  
    // Output information on the sample values in infeasible blocks.
    sprintf(file_name_buffer, "%s.infeasible.txt", argv[batch_index]);
    infeasible_block_information_output(downsampled_chroma_estimate, refined_block_data, file_name_buffer);
  
    // Now replace the data in blocks which are infeasible with simply re-DCTed data.
    substitute_fdct_infeasible_blocks(downsampled_chroma_estimate, refined_block_data, *luma_table, *chroma_table);
  
    // Output the block summaries.
    printf("QF: %d\n", determined_qf);
    printf("Y: %d exact %d ambiguous %d infeasible %d impossible\n", y_block_status.exact_block_count, y_block_status.ambiguous_block_count, y_block_status.infeasible_block_count, y_block_status.impossible_block_count);
    printf("Cb: %d exact %d ambiguous %d infeasible %d impossible\n", cb_block_status.exact_block_count, cb_block_status.ambiguous_block_count, cb_block_status.infeasible_block_count, cb_block_status.impossible_block_count);
    printf("Cr: %d exact %d ambiguous %d infeasible %d impossible\n", cr_block_status.exact_block_count, cr_block_status.ambiguous_block_count, cr_block_status.infeasible_block_count, cr_block_status.impossible_block_count);

    // Flush the output.
    fflush(stdout);
  
    // Destroy the quantization tables.
    destroy_quantization_table(*luma_table);
    destroy_quantization_table(*chroma_table);
  
    // TODO: This will fail if quantization was impossible, as refined_block_data will be NULL.
    // Output a .jpg file with the recompressed data.
    sprintf(file_name_buffer, "%s.result.jpg", argv[batch_index]);
    jpeg_output(refined_block_data, determined_qf, file_name_buffer);
  
    progress_output("Cleaning up data structures\n");
    free(luma_table);
    free(chroma_table);
    destroy_set_plane(upsampled_component_planes);
    destroy_packed_image_data(image);
    destroy_computation_plane(smoothing_computations);
    destroy_block_interval_component_image_data(refined_block_data);
    destroy_interval_component_image_data(downsampled_chroma_estimate);
  }

  destroy_inverse_ycc_rgb_table(inverse_ycc_rgb_table);
  free(file_name_buffer);

  return 0;
}

// Printf for progress output.
void progress_output(const char* fstring, ...)
{
  va_list args;
  va_start(args, fstring);
#ifdef PROGRESS_OUTPUT
  vprintf(fstring, args);
#endif
  va_end(args);
}

// Read a binary 24-bit PPM image.
packed_image_data load_rgb_ppm(const char* file_name)
{
  FILE* file = fopen(file_name, "rb");
  char* buffer = (char*)malloc(READ_BUFFER_SIZE * sizeof(char));
  UINT16 width, height;

  // Image type.
  if (!fgets(buffer, sizeof(char) * READ_BUFFER_SIZE, file))
    printf("rjpeg.c: Error reading input file.\n");
  if (buffer[1] != '6')
    printf("rjpeg.c: The input PPM must be in P6 mode. Trying to proceed anyway.\n");

  // Image dimensions.
  buffer[0] = '#';
  while (buffer[0] == '#')
    if (!fgets(buffer, sizeof(char) * READ_BUFFER_SIZE, file))
      printf("rjpeg.c: Error reading input file.\n");
  sscanf(buffer, "%u %u", &width, &height);

  // Sample maximum value (assume 255).
  if (!fgets(buffer, sizeof(char) * READ_BUFFER_SIZE, file))
    printf("rjpeg.c: Error reading input file.\n");
  
  // Read the image data into a new RGB image data structure.
  packed_image_data result = create_packed_image_data(width, height);
  result.samples = (unsigned char*)malloc(result.width * result.height * 3 * sizeof(unsigned char));
  if (!fread(result.samples, sizeof(unsigned char), result.width * result.height * 3, file))
    printf("rjpeg.c: Error reading input file.\n");
  fclose(file);
  free(buffer);

  return result;
}

// Write a greyscale 8-bit PPM image.
void write_grey_ppm(char* file_name, interval_plane values)
{
  unsigned char* byte_values = (unsigned char*)malloc(values.width * values.height * sizeof(unsigned char));
  FILE* file = fopen(file_name, "wb");
  int x, y;
  for (y = 0; y < values.height; y++)
    for (x = 0; x < values.width; x++)
      byte_values[x + y * values.width] = (unsigned char)values.min[x + y * values.width];
  fprintf(file, "P5\n%d %d\n255\n", values.width, values.height);
  fwrite(byte_values, sizeof(unsigned char), values.width * values.height, file);
  fclose(file);
  free(byte_values);
}

// Write a binary 24-bit PPM image.
void write_rgb_ppm(char* file_name, packed_image_data image)
{
  FILE* file = fopen(file_name, "wb");
  fprintf(file, "P6\n%d %d\n255\n", image.width, image.height);
  fwrite(image.samples, sizeof(unsigned char), image.width * image.height * 3, file);
  fclose(file);
}
