// Copyright (c) 2010 Andrew Lewis. All rights reserved.
// Use of this source code is governed the license in the LICENSE file.

#include "solver.h"

// This function is not normally used, but might be useful for diagnostics.
UINT8* load_grey(const char* file_name)
{
  FILE* file = fopen(file_name, "rb");
  char* buffer = (char*)malloc(256 * sizeof(char));
  UINT16 width, height;

  // Image type.
  if (!fgets(buffer, sizeof(char) * 256, file))
    printf("solver.c: Error reading grey data.\n");
  if (buffer[1] != '5')
    printf("The input PPM must be in P5 mode. Trying to proceed anyway.\n");

  // Image dimensions.
  buffer[0] = '#';
  while (buffer[0] == '#')
    if (!fgets(buffer, sizeof(char) * 256, file))
      printf("solver.c: Error reading grey data.\n");
  sscanf(buffer, "%u %u", &width, &height);

  // Sample maximum value (assume 255).
  if (!fgets(buffer, sizeof(char) * 256, file))
    printf("solver.c: Error reading grey data.\n");
  
  // Read the image data into a new RGB image data structure.
  UINT8* samples = (unsigned char*)malloc(width * height * 3 * sizeof(unsigned char));
  if (!fread(samples, sizeof(unsigned char), width * height * 3, file))
    printf("solver.c: Error reading grey data.\n");
  fclose(file);
  free(buffer);

  return samples;
}

void calculate_input_intervals(computation_plane computations, interval_plane inputs, set_plane outputs, UINT8 colour_shift)
{
  UINT16 x, y, ux, uy;
  UINT16 i, set_index;
  INT8 dx, dy;
  INT32 singleton_count;
  UINT64 total_imprecision, old_imprecision;
  INT64 oMax, oMin;
  interval_plane union_buffer;

  // Set up the union buffer.
  union_buffer = create_interval_plane(inputs.width, inputs.height);
  for (uy = 0; uy < inputs.height; uy++)
  {
    for (ux = 0; ux < inputs.width; ux++)
    {
      union_buffer.min[uy * inputs.width + ux] = union_buffer.max[uy * inputs.width + ux] = UNION_BUFFER_UNDEFINED;
      inputs.min[uy * inputs.width + ux] = 0;
      inputs.max[uy * inputs.width + ux] = 255;
    }
  }

  // Iteratively reduce the imprecision until it is unchanged.
  i = 0;
  dx = dy = 1;
  total_imprecision = 0;
  do
  {
    old_imprecision = total_imprecision;

    // Make a pass over the intervals updating them. The direction of the pass is determined by dx, dy.
    for (y = (dy == -1 ? outputs.height - 1 : 0); y < outputs.height; y += dy)
    {
      for (x = (dx == -1 ? outputs.width - 1 : 0); x < outputs.width; x += dx)
      {
        // Extract each element in the set for this output, reverse the computations and combine the information into the union buffer.
        for (set_index = 1; set_index < outputs.sets[y * outputs.width + x][0]; set_index++)
        {
          if (outputs.sets[y * outputs.width + x][set_index] == LIST_SENTINEL32)
            break;
          oMax = oMin = ((outputs.sets[y * outputs.width + x][set_index] >> colour_shift) & 0x0000FF);

          reverse_computation(&oMax, &oMin, computations.computations[y * outputs.width + x], union_buffer, TRUE, inputs, FALSE);
        }

        // Intersect the results in the union buffer with the current estimate. Offset determined from filter length.
        for (uy = MAX((INT32)y / 2 - 2, 0); uy < MIN((INT32)y / 2 + 2, inputs.height); uy++)
        {
          for (ux = MAX((INT32)x / 2 - 2, 0); ux < MIN((INT32)x / 2 + 2, inputs.width); ux++)
          {
            if (union_buffer.min[uy * inputs.width + ux] != UNION_BUFFER_UNDEFINED || union_buffer.max[uy * inputs.width + ux] != UNION_BUFFER_UNDEFINED)
            {
              if (union_buffer.min[uy * inputs.width + ux] > inputs.min[uy * inputs.width + ux])
              {
                inputs.min[uy * inputs.width + ux] = union_buffer.min[uy * inputs.width + ux];
              }
              if (union_buffer.max[uy * inputs.width + ux] < inputs.max[uy * inputs.width + ux])
              {
                inputs.max[uy * inputs.width + ux] = union_buffer.max[uy * inputs.width + ux];
              }

              if (inputs.min[uy * inputs.width + ux] > inputs.max[uy * inputs.width + ux])
                printf("  Wrong %d %d at %d %d x y %d %d\n", inputs.min[uy * inputs.width + ux], inputs.max[uy * inputs.width + ux], ux, uy, x, y);

              // Reset the union buffer.
              union_buffer.min[uy * inputs.width + ux] = union_buffer.max[uy * inputs.width + ux] = UNION_BUFFER_UNDEFINED;
            }
          }
        }
      }
    }

    dx = (dx == -1 ? 1 : -1);
    dy = (dx == 1 ? (dy == -1 ? 1 : -1) : dy);

    // Sum up the imprecision. If it stops changing, we have reached a fixed point.
    total_imprecision = 0;
    singleton_count = 0;
    for (y = 0; y < inputs.height; y++)
    {
      for (x = 0; x < inputs.width; x++)
      {
        if (inputs.max[y * inputs.width + x] - inputs.min[y * inputs.width + x] == 0)
          singleton_count++;
        total_imprecision += (inputs.max[y * inputs.width + x] - inputs.min[y * inputs.width + x]);
      }
    }

//    printf("  Umsooth iteration %02d: imprecision %d singletons %d\n", i, total_imprecision, singleton_count);
    i++;
  }
  while (total_imprecision != old_imprecision && total_imprecision != 0);

  // Clean up temporary storage.
  destroy_interval_plane(union_buffer);
}
